;; 
;; FFIX.COM -- Fix up the FWAIT bug in the Microsoft FPU emulation library
;;             used in QuickBASIC 4.x; PDS 6+ and VBDOS 1 interpreters and
;;             compilers.
;;
;; Description:
;;      This program will traps (and deny) any try to set a new address
;; to the interrupt vector 3Dh (through DOS) and generate inline code when
;; this interrupt is invoked.
;;
;; 	This vector is used by the Microsoft floating-point unit emulation 
;; package to mimic the FWAIT instruction. This package _would_ act this way: 
;; every time an interrupt (hooked by it) is invoked, the ISR check if the 
;; computer where it is running has a math coprocessor installed. If the 
;; answer is positive, it _must_ substitute the INT XX instruction that had 
;; invoked it by the equivalent FPU instruction. 
;;
;; 	But with FWAIT (INT 3Dh) this doesn't occur. The interrupt service 
;; routine will never generate inline code even knowing that there is a 
;; coprocessor installed.
;;
;; 	As an INT instruction is very time consuming (much more if the 
;; program is running in Virtual86 mode, i.e.: a DOS box in Windows) and as 
;; BASIC (mainly QB4.x) use _a lot_ of FWAITs when doing floating-point 
;; arithmetics, you can imagine how the speed will drops. In my computer the 
;; difference between a program (that use a lot of SINGLEs and DOUBLE 
;; variables in its calculations) running with the bug and the same program 
;; running with FFIX loaded is like 300%.
;;
;; 	I have discovered this bug in 98 when _narf posted an message saying 
;; that his assembly functions (those included in the GSLIB) were much more 
;; slower than the QB ones. So, as i didn't have nothing better to do :}, i 
;; did some tests and as the same thing happened, i did run the tests one 
;; more time using SoftIce (wonderful tool :]) to debug it and then...the 
;; bug smacked my face :). So, i made a fix up to it and _narf added it to 
;; his library.
;;
;; Resuming:
;;      Execute this program and run some BAS programs (compiled or through 
;; the IDE) that are floating-point intensive to see the speed up!
;;
;;
;; copyleft 2000 by v1ctor [av1ctor@yahoo.com.br]
;;
;; make it: tasm /m ffix
;;          tlink /t/x ffix
;;      or: masm ffix _or_ ml /c ffix.asm
;;          link /t ffix,,nul;
;;      or: a86 +s ffix.asm

                .model  tiny                    ;; cs=ds=es=ss=PSP
                .186                            ;; for the shift(s)...

                ;; some short-cuts
                O               equ offset
                W               equ word ptr    ;; comment this line with a86

                ;; some DOS services
                DOS             equ 21h
                DOS_MSG_SHOW    equ 09h
                DOS_INT_SET     equ 25h
                DOS_TSR         equ 31h
                DOS_INT_GET     equ 35h
                DOS_MEM_FREE    equ 49h
                DOS_EXIT        equ 4Ch
               
.code
                org     100h                    ;; skip PSP
;;:: resident part :::::::::::::::::::::::::::::::::::::::::::::::::::::::::
start:          jmp     startup

org_dos_int     dd      ?                       ;; DON'T MOVE!
current_PSP     dw      ?                       ;; /

;;::::::::::::::
                db      'FFIX'                  ;; signature (DON'T MOVE!)
new_int_3d:     push    bp
                mov     bp, sp
                push    ds

                lds     bp, [bp + 2]            ;; ds:bp -> return address
                mov     W ds:[bp-2], 909Bh      ;; fix with: fwait; nop

                pop     ds
                pop     bp
                iret

;;::::::::::::::
                db      'FFIX'                  ;; same as above
new_int_21:     push    ax
                push    dx

                mov     dx, ax
                lahf                            ;; save flags

                cmp     dh, DOS_INT_SET
                je      @@check_vector          ;; trying set a int vector?

@@chain:        sahf                            ;; restore flags
                pop     dx
                pop     ax
                jmp     dword ptr cs:org_dos_int;; chain to DOS

@@check_vector: cmp     dl, 3Dh
                jne     @@chain                 ;; vector 3Dh?

                ;; deny it...
                pop     dx
                pop     ax
                iret

;;:: transient part ::::::::::::::::::::::::::::::::::::::::::::::::::::::::
startup:        ;; show intro message
                mov     dx, O msg_intro
                mov     ah, DOS_MSG_SHOW
                int     DOS

                ;; check command line for option: /u
                mov     bx, 81h
@@loop:         mov     al, ds:[bx]
                inc     bx
                cmp     al, 13
                je      @@no_cmd
                cmp     al, 32
                je      @@loop
                cmp     al, 9
                je      @@loop
                cmp     al, '/'
                jne     @@opt_show
                mov     al, ds:[bx]
                or      al, 32                  ;; to lower case
                cmp     al, 'u'
                jne     @@opt_show

;;..............
                ;; check if previously loaded
                mov     ax, (DOS_INT_GET*256) + 3Dh
                int     DOS
                cmp     W es:[bx-4], 'FF'
                jne     @@not_loaded            ;; signature= FFIX?
                cmp     W es:[bx-2], 'XI'
                jne     @@not_loaded            ;; /

                ;; check if we can unload it
                push    bx                      ;; save
                push    es                      ;; /
                mov     ax, (DOS_INT_GET*256) + DOS
                int     DOS
                cmp     W es:[bx-4], 'FF'
                jne     @@no_unload             ;; signature= FFIX?
                cmp     W es:[bx-2], 'XI'
                jne     @@no_unload             ;; /

                ;; yep, we can...
                pop     es                      ;; restore
                pop     bx                      ;; /

                ;; restore original DOS interrupt vector
                push    ds
                lds     dx, es:[bx-4-2-4]       ;; ds:dx -> org DOS int
                mov     ax, (DOS_INT_SET*256) + DOS
                int     DOS

                ;; set int vector 3Dh to NULL
                xor     dx, dx
                mov     ds, dx
                mov     ax, (DOS_INT_SET*256) + 3Dh
                int     DOS
                pop     ds

                ;; free mem occupied by the previews loaded FFIX's PSP+code
                mov     es, es:[bx-4-2]
                mov     ah, DOS_MEM_FREE
                int     DOS

                mov     dx, O msg_unloaded
                jmp     short @@exit
                
@@no_unload:    add     sp, 2 + 2               ;; free stack
                mov     dx, O msg_cant_unload
                jmp     short @@exit_err

@@not_loaded:   mov     dx, O msg_not_loaded
                jmp     short @@exit_err

@@opt_show:     mov     dx, O msg_options

@@exit:         mov     ah, DOS_MSG_SHOW
                int     DOS

                mov     ax, (DOS_EXIT*256) + 00h
                int     DOS
                
@@no_cmd:       ;; check if already loaded
                mov     ax, (DOS_INT_GET*256) + 3Dh
                int     DOS
                cmp     W es:[bx-4], 'FF'
                jne     @@load                  ;; signature= FFIX?
                cmp     W es:[bx-2], 'XI'
                jne     @@load                  ;; /

                ;; yep, so get out...
                mov     dx, O msg_already

@@exit_err:     mov     ah, DOS_MSG_SHOW
                int     DOS

                ;; exit with errorlevel 1
                mov     ax, (DOS_EXIT*256) + 01h
                int     DOS

;;..............
@@load:         mov     cs:current_PSP, cs      ;; save PSP to unload later

                ;; get current DOS interrupt vector
                mov     ax, (DOS_INT_GET*256) + DOS
                int     DOS
                mov     W cs:org_dos_int, bx
                mov     W cs:org_dos_int+2, es

                ;; put new_int_3d on IVT
                mov     dx, O new_int_3d        ;; ds:dx -> new_int_3d
                mov     ax, (DOS_INT_SET*256) + 3Dh
                int     DOS

                ;; and new_int_21 too
                mov     dx, O new_int_21        ;; ds:dx -> new_int_21
                mov     ax, (DOS_INT_SET*256) + DOS
                int     DOS

                ;; show 'loaded' message
                mov     dx, O msg_loaded
                mov     ah, DOS_MSG_SHOW
                int     DOS

                ;; free environment block; address (seg) at PSP:2Ch
                mov     es, ds:[2Ch]
                mov     ah, DOS_MEM_FREE
                int     DOS

                ;; terminate & stay resident -- from PSP to `startup'
                mov     dx, O startup + 15
                shr     dx, 4                   ;; dx= bytes 2 keep (in paras)
                mov     ax, (DOS_TSR*256) + 00h
                int     DOS

                ;; that's all, easy not? ;)

msg_intro       db      'FFIX (c) 2000 by v1ctor [av1ctor@yahoo.com.br]',13,10,'$'
msg_loaded      db      'loaded to memory.', 13,10, '$'
msg_already     db      'already loaded!', 13,10, '$'
msg_not_loaded  db      'not loaded!', 13,10, '$'
msg_cant_unload db      'cannot unload!', 13,10, '$'
msg_unloaded    db      'unloaded from memory.', 13,10, '$'
msg_options     db      'to load  : ffix',13,10, 'to unload: ffix /u',13,10,'$'
                end     start
